@@ -16,20 +16,14 @@
 #include <llvm/IR/DataLayout.h>
 #include <llvm/IR/Mangler.h>
 #include <llvm/ExecutionEngine/RuntimeDyld.h>
-#if JL_LLVM_VERSION >= 50000
 #include <llvm/BinaryFormat/Magic.h>
-#endif
 #include <llvm/Object/MachO.h>
 #include <llvm/Object/COFF.h>
 #include <llvm/Object/ELFObjectFile.h>
 
 using namespace llvm;
 
-#if JL_LLVM_VERSION >= 50000
 using llvm_file_magic = file_magic;
-#else
-using llvm_file_magic = sys::fs::file_magic;
-#endif
 
 #include "julia.h"
 #include "julia_internal.h"
@@ -404,11 +398,7 @@ public:
                 ObjectInfo tmp = {&debugObj,
                     (size_t)SectionSize,
                     (ptrdiff_t)(SectionAddr - SectionLoadAddr),
-#if JL_LLVM_VERSION >= 60000
                     DWARFContext::create(debugObj, &L).release(),
-#else
-                    new DWARFContextInMemory(debugObj, &L),
-#endif
                     };
                 objectmap[SectionLoadAddr] = tmp;
                 first = false;
@@ -948,11 +938,7 @@ static objfileentry_t &find_object_file(uint64_t fbase, StringRef fname)
             slide = -(int64_t)fbase;
         }
 
-#if JL_LLVM_VERSION >= 60000
         auto context = DWARFContext::create(*debugobj).release();
-#else
-        auto context = new DWARFContextInMemory(*debugobj);
-#endif
         auto binary = errorobj->takeBinary();
         binary.first.release();
         binary.second.release();
