@@ -274,15 +274,19 @@ static void ti_initthread(int16_t tid)
                                     sizeof(size_t));
     }
     ptls->defer_signal = 0;
-    ptls->current_module = NULL;
     void *bt_data = malloc(sizeof(uintptr_t) * (JL_MAX_BT_SIZE + 1));
-    memset(bt_data, 0, sizeof(uintptr_t) * (JL_MAX_BT_SIZE + 1));
     if (bt_data == NULL) {
         jl_printf(JL_STDERR, "could not allocate backtrace buffer\n");
         gc_debug_critical_error();
         abort();
     }
+    memset(bt_data, 0, sizeof(uintptr_t) * (JL_MAX_BT_SIZE + 1));
     ptls->bt_data = (uintptr_t*)bt_data;
+    ptls->sig_exception = NULL;
+    ptls->previous_exception = NULL;
+#ifdef _OS_WINDOWS_
+    ptls->needs_resetstkoflw = 0;
+#endif
     jl_init_thread_heap(ptls);
     jl_install_thread_signal_handler(ptls);
 
@@ -321,7 +325,7 @@ static jl_value_t *ti_run_fun(jl_callptr_t fptr, jl_method_instance_t *mfunc,
             ptls->safe_restore = &buf;
             jl_printf(JL_STDERR, "\nError thrown in threaded loop on thread %d: ",
                       (int)ptls->tid);
-            jl_static_show(JL_STDERR, ptls->exception_in_transit);
+            jl_static_show(JL_STDERR, jl_current_exception());
         }
         ptls->safe_restore = old_buf;
         JL_UNLOCK_NOGC(&lock);
@@ -361,13 +365,11 @@ void ti_threadfun(void *arg)
 
     // initialize this thread (set tid, create heap, etc.)
     ti_initthread(ta->tid);
-    jl_init_stack_limits(0);
+    void *stack_lo, *stack_hi;
+    jl_init_stack_limits(0, &stack_lo, &stack_hi);
 
     // set up tasking
-    jl_init_root_task(ptls->stack_lo, ptls->stack_hi - ptls->stack_lo);
-#ifdef COPY_STACKS
-    jl_set_base_ctx((char*)&arg);
-#endif
+    jl_init_root_task(stack_lo, stack_hi);
 
     // set the thread-local tid and wait for a thread group
     while (jl_atomic_load_acquire(&ta->state) == TI_THREAD_INIT)
@@ -396,6 +398,7 @@ void ti_threadfun(void *arg)
 
         ti_threadgroup_fork(tg, ptls->tid, (void **)&work, init);
         init = 0;
+        JL_GC_PROMISE_ROOTED(work);
 
 #if PROFILE_JL_THREADING
         uint64_t tfork = uv_hrtime();
@@ -414,15 +417,10 @@ void ti_threadfun(void *arg)
                 //       enter GC unsafe region when starting the work.
                 int8_t gc_state = jl_gc_unsafe_enter(ptls);
                 // This is probably always NULL for now
-                jl_module_t *last_m = ptls->current_module;
                 size_t last_age = ptls->world_age;
-                JL_GC_PUSH1(&last_m);
-                ptls->current_module = work->current_module;
                 ptls->world_age = work->world_age;
                 ti_run_fun(work->fptr, work->mfunc, work->args, work->nargs);
-                ptls->current_module = last_m;
                 ptls->world_age = last_age;
-                JL_GC_POP();
                 jl_gc_unsafe_leave(ptls, gc_state);
             }
         }
@@ -609,12 +607,15 @@ void jl_start_threads(void)
         mask[0] = 0;
     }
 
+    // The analyzer doesn't know jl_n_threads doesn't change, help it
+    size_t nthreads = jl_n_threads;
+
     // create threads
-    targs = (ti_threadarg_t **)malloc((jl_n_threads - 1) * sizeof (ti_threadarg_t *));
+    targs = (ti_threadarg_t **)malloc((nthreads - 1) * sizeof (ti_threadarg_t *));
 
-    uv_barrier_init(&thread_init_done, jl_n_threads);
+    uv_barrier_init(&thread_init_done, nthreads);
 
-    for (i = 0;  i < jl_n_threads - 1;  ++i) {
+    for (i = 0;  i < nthreads - 1;  ++i) {
         targs[i] = (ti_threadarg_t *)malloc(sizeof (ti_threadarg_t));
         targs[i]->state = TI_THREAD_INIT;
         targs[i]->tid = i + 1;
@@ -628,13 +629,13 @@ void jl_start_threads(void)
     }
 
     // set up the world thread group
-    ti_threadgroup_create(1, jl_n_threads, 1, &tgworld);
-    for (i = 0;  i < jl_n_threads;  ++i)
+    ti_threadgroup_create(1, nthreads, 1, &tgworld);
+    for (i = 0;  i < nthreads;  ++i)
         ti_threadgroup_addthread(tgworld, i, NULL);
     ti_threadgroup_initthread(tgworld, ptls->tid);
 
     // give the threads the world thread group; they will block waiting for fork
-    for (i = 0;  i < jl_n_threads - 1;  ++i) {
+    for (i = 0;  i < nthreads - 1;  ++i) {
         targs[i]->tg = tgworld;
         jl_atomic_store_release(&targs[i]->state, TI_THREAD_WORK);
     }
@@ -701,7 +702,6 @@ JL_DLLEXPORT jl_value_t *jl_threading_run(jl_value_t *_args)
     threadwork.args = args;
     threadwork.nargs = nargs;
     threadwork.ret = jl_nothing;
-    threadwork.current_module = ptls->current_module;
     threadwork.world_age = world;
 
 #if PROFILE_JL_THREADING
@@ -719,6 +719,7 @@ JL_DLLEXPORT jl_value_t *jl_threading_run(jl_value_t *_args)
 #endif
 
     // this thread must do work too (TODO: reduction?)
+    JL_GC_PROMISE_ROOTED(threadwork.mfunc);
     tw->ret = ti_run_fun(threadwork.fptr, threadwork.mfunc, args, nargs);
 
 #if PROFILE_JL_THREADING
