@@ -91,18 +91,6 @@ end
 
 (*)(x::P, y::Real) where {P<:Period} = P(value(x) * Int64(y))
 (*)(y::Real, x::Period) = x * y
-for (op, Ty, Tz) in ((:*, Real, :P),
-                   (:/, :P, Float64), (:/, Real, :P))
-    @eval begin
-        function ($op)(X::StridedArray{P}, y::$Ty) where P<:Period
-            Z = similar(X, $Tz)
-            for (Idst, Isrc) in zip(eachindex(Z), eachindex(X))
-                @inbounds Z[Idst] = ($op)(X[Isrc], y)
-            end
-            return Z
-        end
-    end
-end
 
 # intfuncs
 Base.gcdx(a::T, b::T) where {T<:Period} = ((g, x, y) = gcdx(value(a), value(b)); return T(g), x, y)
@@ -361,14 +349,6 @@ Base.show(io::IO,x::CompoundPeriod) = print(io, string(x))
 
 GeneralPeriod = Union{Period, CompoundPeriod}
 (+)(x::GeneralPeriod) = x
-(+)(x::StridedArray{<:GeneralPeriod}) = x
-
-for op in (:+, :-)
-    @eval begin
-        ($op)(X::StridedArray{<:GeneralPeriod}, Y::StridedArray{<:GeneralPeriod}) =
-            reshape(CompoundPeriod[($op)(x, y) for (x, y) in zip(X, Y)], promote_shape(size(X), size(Y)))
-    end
-end
 
 (==)(x::CompoundPeriod, y::Period) = x == CompoundPeriod(y)
 (==)(x::Period, y::CompoundPeriod) = y == x
