@@ -78,7 +78,7 @@ julia> iA = [4 3; 6 3]
  6  3
 
 julia> lu!(iA)
-ERROR: InexactError: Int64(Int64, 0.6666666666666666)
+ERROR: InexactError: Int64(0.6666666666666666)
 Stacktrace:
 [...]
 ```
@@ -305,7 +305,7 @@ issuccess(F::LU) = F.info == 0
 
 function show(io::IO, mime::MIME{Symbol("text/plain")}, F::LU)
     if issuccess(F)
-        println(io, summary(F))
+        summary(io, F); println(io)
         println(io, "L factor:")
         show(io, mime, F.L)
         println(io, "\nU factor:")
