@@ -6,6 +6,7 @@ using .Main.OffsetArrays
 using SparseArrays
 
 using Random, LinearAlgebra
+using Dates
 
 @testset "basics" begin
     @test length([1, 2, 3]) == 3
@@ -234,6 +235,29 @@ end
     end
 end
 
+@testset "reshape isbitsunion Arrays (issue #28611)" begin
+    v = Union{Float64,Missing}[]
+    for i in 1:10 push!(v, i) end
+    v[5] = missing
+    a = @inferred(reshape(v, 2, 5))
+    for (I, i) in zip(CartesianIndices((2, 5)), 1:10)
+        @test a[I] === v[i]
+    end
+    ac = copy(a)
+    @test ac isa Array{Union{Float64, Missing}, 2}
+    b = @inferred(reshape(ac, 5, 2))
+    for (I, J) in zip(CartesianIndices((2, 5)), CartesianIndices((5, 2)))
+        @test ac[I] === b[J]
+    end
+
+    for T in (Any, Union{Float64,String})
+        a = Array{T}(undef, 4)
+        @test @inferred(reshape(a, (2, 2))) isa Array{T,2}
+        a = Array{T}(undef, 4, 1)
+        @test @inferred(reshape(a, (2, 2))) isa Array{T,2}
+    end
+end
+
 @testset "conversion from ReshapedArray to Array (#18262)" begin
     a = Base.ReshapedArray(1:3, (3, 1), ())
     @test convert(Array, a) == a
@@ -319,20 +343,27 @@ end
     @test B == [0 23 1 24 0; 11 12 13 14 15; 0 21 3 22 0; 0 7 7 0 0]
 
     @test isequal(reshape(reshape(1:27, 3, 3, 3), Val(2))[1,:], [1,  4,  7,  10,  13,  16,  19,  22,  25])
-
+end
+@testset "find(in(b), a)" begin
+    # unsorted inputs
     a = [3, 5, -7, 6]
     b = [4, 6, 2, -7, 1]
-    ind = findall(in(b), a)
-    @test ind == [3,4]
+    @test findall(in(b), a) == [3,4]
     @test findall(in(Int[]), a) == Int[]
     @test findall(in(a), Int[]) == Int[]
+    @test findall(in(b), reshape(a, 2, 2)) == [CartesianIndex(1, 2), CartesianIndex(2, 2)]
 
-    a = [1,2,3,4,5]
+    # sorted inputs
+    a = [1,2,3,4,5,10]
     b = [2,3,4,6]
     @test findall(in(b), a) == [2,3,4]
     @test findall(in(a), b) == [1,2,3]
     @test findall(in(Int[]), a) == Int[]
     @test findall(in(a), Int[]) == Int[]
+    @test findall(in(b), reshape(a, 3, 2)) ==
+        [CartesianIndex(2, 1), CartesianIndex(3, 1), CartesianIndex(1, 2)]
+    @test findall(in(a), reshape(b, 2, 2)) ==
+        [CartesianIndex(1, 1), CartesianIndex(2, 1), CartesianIndex(1, 2)]
 
     a = Vector(1:3:15)
     b = Vector(2:4:10)
@@ -350,7 +381,8 @@ end
 
     @test findall(in([1, 2]), 2) == [1]
     @test findall(in([1, 2]), 3) == []
-
+end
+@testset "setindex! return type" begin
     rt = Base.return_types(setindex!, Tuple{Array{Int32, 3}, Vector{UInt8}, Vector{Int}, Int16, UnitRange{Int}})
     @test length(rt) == 1 && rt[1] === Array{Int32, 3}
 end
@@ -567,6 +599,11 @@ end
     @test isnan(findmax([NaN, NaN, 0.0/0.0])[1])
     @test findmax([NaN, NaN, 0.0/0.0])[2] == 1
 
+    # Check that cartesian indices are returned for matrices
+    @test argmax([10 12; 9 11]) === CartesianIndex(1, 2)
+    @test argmin([10 12; 9 11]) === CartesianIndex(2, 1)
+    @test findmax([10 12; 9 11]) === (12, CartesianIndex(1, 2))
+    @test findmin([10 12; 9 11]) === (9, CartesianIndex(2, 1))
 end
 
 @testset "permutedims" begin
@@ -699,6 +736,12 @@ end
     @test_throws MethodError repeat(1, 2, 3)
     @test repeat([1, 2], 1, 2, 3) == repeat([1, 2], outer = (1, 2, 3))
 
+    # issue 29614
+    @test repeat(ones(2, 2), 1, 1, 1) == ones(2, 2, 1)
+    @test repeat(ones(2, 2), 2, 2, 2) == ones(4, 4, 2)
+    @test repeat(ones(2), 2, 2, 2) == ones(4, 2, 2)
+    @test repeat(ones(2, 2), inner=(1, 1, 1), outer=(2, 2, 2)) == ones(4, 4, 2)
+
     R = repeat([1, 2])
     @test R == [1, 2]
     R = repeat([1, 2], inner=1)
@@ -1199,11 +1242,13 @@ end
     @test issorted(as[:,1])
     @test issorted(as[:,2])
     @test issorted(as[:,3])
+    @test sort!(copy(a), dims=1) == as
 
     as = sort(a, dims=2)
     @test issorted(as[1,:])
     @test issorted(as[2,:])
     @test issorted(as[3,:])
+    @test sort!(copy(a), dims=2) == as
 
     local b = rand(21,21,2)
 
@@ -1212,15 +1257,18 @@ end
         @test issorted(bs[:,i,1])
         @test issorted(bs[:,i,2])
     end
+    @test sort!(copy(b), dims=1) == bs
 
     bs = sort(b, dims=2)
     for i in 1:21
         @test issorted(bs[i,:,1])
         @test issorted(bs[i,:,2])
     end
+    @test sort!(copy(b), dims=2) == bs
 
     bs = sort(b, dims=3)
     @test all(bs[:,:,1] .<= bs[:,:,2])
+    @test sort!(copy(b), dims=3) == bs
 end
 
 @testset "higher dimensional sortslices" begin
@@ -1323,6 +1371,12 @@ end
     @test_throws BoundsError deleteat!(a, Bool[])
     @test_throws BoundsError deleteat!(a, [true])
     @test_throws BoundsError deleteat!(a, falses(11))
+
+    @test_throws BoundsError deleteat!([], 1)
+    @test_throws BoundsError deleteat!([], [1])
+    @test_throws BoundsError deleteat!([], [2])
+    @test deleteat!([], []) == []
+    @test deleteat!([], Bool[]) == []
 end
 
 @testset "comprehensions" begin
@@ -1686,13 +1740,13 @@ end
     @test I2 + I1 == CartesianIndex((1,8,2))
     @test I1 - I2 == CartesianIndex((3,-2,-2))
     @test I2 - I1 == CartesianIndex((-3,2,2))
-    @test I1 + 1*one(I1) == CartesianIndex((3,4,1))
-    @test I1 - 2*one(I1) == CartesianIndex((0,1,-2))
+    @test I1 + 1*oneunit(I1) == CartesianIndex((3,4,1))
+    @test I1 - 2*oneunit(I1) == CartesianIndex((0,1,-2))
 
     @test zero(CartesianIndex{2}) == CartesianIndex((0,0))
     @test zero(CartesianIndex((2,3))) == CartesianIndex((0,0))
-    @test one(CartesianIndex{2}) == CartesianIndex((1,1))
-    @test one(CartesianIndex((2,3))) == CartesianIndex((1,1))
+    @test oneunit(CartesianIndex{2}) == CartesianIndex((1,1))
+    @test oneunit(CartesianIndex((2,3))) == CartesianIndex((1,1))
 
     @test min(CartesianIndex((2,3)), CartesianIndex((5,2))) == CartesianIndex((2,2))
     @test max(CartesianIndex((2,3)), CartesianIndex((5,2))) == CartesianIndex((5,3))
@@ -1723,8 +1777,12 @@ end
     b[CartesianIndex{2}(1,1)] = 7
     @test a[1,2] == 7
     @test 2*CartesianIndex{3}(1,2,3) == CartesianIndex{3}(2,4,6)
+    @test CartesianIndex{3}(1,2,3)*2 == CartesianIndex{3}(2,4,6)
+    @test_throws ErrorException iterate(CartesianIndex{3}(1,2,3))
+    @test CartesianIndices(CartesianIndex{3}(1,2,3)) == CartesianIndices((1, 2, 3))
+    @test Tuple{}(CartesianIndices{0,Tuple{}}(())) == ()
 
-    R = CartesianIndices(map(Base.Slice, (2:5, 3:5)))
+    R = CartesianIndices(map(Base.IdentityUnitRange, (2:5, 3:5)))
     @test eltype(R) <: CartesianIndex{2}
     @test eltype(typeof(R)) <: CartesianIndex{2}
     @test eltype(CartesianIndices{2}) <: CartesianIndex{2}
@@ -1748,6 +1806,10 @@ end
 
     @test @inferred(convert(NTuple{2,UnitRange}, R)) === (2:5, 3:5)
     @test @inferred(convert(Tuple{Vararg{UnitRange}}, R)) === (2:5, 3:5)
+
+    I = CartesianIndex(2,3)
+    J = CartesianIndex(5,4)
+    @test I:J === CartesianIndices((2:5, 3:4))
 end
 
 # All we really care about is that we have an optimized
@@ -1923,6 +1985,22 @@ end
     @test IndexStyle(selectdim(A, 3, 1)) == IndexStyle(view(A, :, :, 1)) == IndexLinear()
 end
 
+# row/column/slice iterator tests
+using Base: eachrow, eachcol
+@testset "row/column/slice iterators" begin
+    # Simple ones
+    M = [1 2 3; 4 5 6; 7 8 9]
+    @test collect(eachrow(M)) == collect(eachslice(M, dims = 1)) == [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
+    @test collect(eachcol(M)) == collect(eachslice(M, dims = 2)) == [[1, 4, 7], [2, 5, 8], [3, 6, 9]]
+    @test_throws DimensionMismatch eachslice(M, dims = 4)
+
+    # Higher-dimensional case
+    M = reshape([(1:16)...], 2, 2, 2, 2)
+    @test_throws MethodError collect(eachrow(M))
+    @test_throws MethodError collect(eachcol(M))
+    @test collect(eachslice(M, dims = 1))[1][:, :, 1] == [1 5; 3 7]
+end
+
 ###
 ### IndexCartesian workout
 ###
@@ -2090,38 +2168,38 @@ let f = OOB_Functor([1,2])
     @test_throws BoundsError map(f, [1,2,3,4,5])
 end
 
-# issue 15654
-@test cumprod([5], dims=2) == [5]
-@test cumprod([1 2; 3 4], dims=3) == [1 2; 3 4]
-@test cumprod([1 2; 3 4], dims=1) == [1 2; 3 8]
-@test cumprod([1 2; 3 4], dims=2) == [1 2; 3 12]
+@testset "issue 15654" begin
+    @test cumprod([5], dims=2) == [5]
+    @test cumprod([1 2; 3 4], dims=3) == [1 2; 3 4]
+    @test cumprod([1 2; 3 4], dims=1) == [1 2; 3 8]
+    @test cumprod([1 2; 3 4], dims=2) == [1 2; 3 12]
 
-@test cumsum([5], dims=2) == [5]
-@test cumsum([1 2; 3 4], dims=1) == [1 2; 4 6]
-@test cumsum([1 2; 3 4], dims=2) == [1 3; 3 7]
-@test cumsum([1 2; 3 4], dims=3) == [1 2; 3 4]
+    @test cumsum([5], dims=2) == [5]
+    @test cumsum([1 2; 3 4], dims=1) == [1 2; 4 6]
+    @test cumsum([1 2; 3 4], dims=2) == [1 3; 3 7]
+    @test cumsum([1 2; 3 4], dims=3) == [1 2; 3 4]
 
-@test cumprod!(Vector{Int}(undef, 1), [5], dims=2) == [5]
-@test cumprod!(Matrix{Int}(undef, 2, 2), [1 2; 3 4], dims=3) == [1 2; 3 4]
-@test cumprod!(Matrix{Int}(undef, 2, 2), [1 2; 3 4], dims=1) == [1 2; 3 8]
-@test cumprod!(Matrix{Int}(undef, 2, 2), [1 2; 3 4], dims=2) == [1 2; 3 12]
+    @test cumprod!(Vector{Int}(undef, 1), [5], dims=2) == [5]
+    @test cumprod!(Matrix{Int}(undef, 2, 2), [1 2; 3 4], dims=3) == [1 2; 3 4]
+    @test cumprod!(Matrix{Int}(undef, 2, 2), [1 2; 3 4], dims=1) == [1 2; 3 8]
+    @test cumprod!(Matrix{Int}(undef, 2, 2), [1 2; 3 4], dims=2) == [1 2; 3 12]
 
-@test cumsum!(Vector{Int}(undef, 1), [5], dims=2) == [5]
-@test cumsum!(Matrix{Int}(undef, 2, 2), [1 2; 3 4], dims=1) == [1 2; 4 6]
-@test cumsum!(Matrix{Int}(undef, 2, 2), [1 2; 3 4], dims=2) == [1 3; 3 7]
-@test cumsum!(Matrix{Int}(undef, 2, 2), [1 2; 3 4], dims=3) == [1 2; 3 4]
-
-# issue #18363
-@test_throws DimensionMismatch cumsum!([0,0], 1:4)
-@test cumsum(Any[])::Vector{Any} == Any[]
-@test cumsum(Any[1, 2.3]) == [1, 3.3] == cumsum(Real[1, 2.3])::Vector{Real}
-@test cumsum([true,true,true]) == [1,2,3]
-@test cumsum(0x00:0xff)[end] === UInt(255*(255+1)÷2) # no overflow
-@test accumulate(+, 0x00:0xff)[end] === 0x80         # overflow
-@test_throws InexactError cumsum!(similar(0x00:0xff), 0x00:0xff) # overflow
-
-@test cumsum([[true], [true], [false]])::Vector{Vector{Int}} == [[1], [2], [2]]
+    @test cumsum!(Vector{Int}(undef, 1), [5], dims=2) == [5]
+    @test cumsum!(Matrix{Int}(undef, 2, 2), [1 2; 3 4], dims=1) == [1 2; 4 6]
+    @test cumsum!(Matrix{Int}(undef, 2, 2), [1 2; 3 4], dims=2) == [1 3; 3 7]
+    @test cumsum!(Matrix{Int}(undef, 2, 2), [1 2; 3 4], dims=3) == [1 2; 3 4]
+end
+@testset "issue #18363" begin
+    @test_throws DimensionMismatch cumsum!([0,0], 1:4)
+    @test cumsum(Any[])::Vector{Any} == Any[]
+    @test cumsum(Any[1, 2.3]) == [1, 3.3] == cumsum(Real[1, 2.3])::Vector{Real}
+    @test cumsum([true,true,true]) == [1,2,3]
+    @test cumsum(0x00:0xff)[end] === UInt(255*(255+1)÷2) # no overflow
+    @test accumulate(+, 0x00:0xff)[end] === 0x80         # overflow
+    @test_throws InexactError cumsum!(similar(0x00:0xff), 0x00:0xff) # overflow
 
+    @test cumsum([[true], [true], [false]])::Vector{Vector{Int}} == [[1], [2], [2]]
+end
 #issue #18336
 @test cumsum([-0.0, -0.0])[1] === cumsum([-0.0, -0.0])[2] === -0.0
 @test cumprod(-0.0im .+ (0:0))[1] === Complex(0.0, -0.0)
@@ -2268,6 +2346,9 @@ end
 
 @testset "diff" begin
     # test diff, throw ArgumentError for invalid dimension argument
+    v = [7, 3, 5, 1, 9]
+    @test diff(v) == [-4, 2, -4, 8]
+    @test diff(v,dims=1) == [-4, 2, -4, 8]
     X = [3  9   5;
          7  4   2;
          2  1  10]
@@ -2277,6 +2358,9 @@ end
     @test diff(view(X, 1:2, 1:2),dims=2) == reshape([6; -3], (2,1))
     @test diff(view(X, 2:3, 2:3),dims=1) == [-3 8]
     @test diff(view(X, 2:3, 2:3),dims=2) == reshape([-2; 9], (2,1))
+    Y = cat([1 3; 4 3], [6 5; 1 4], dims=3)
+    @test diff(Y, dims=3) == reshape([5 2; -3 1], (2, 2, 1))
+    @test_throws UndefKeywordError diff(X)
     @test_throws ArgumentError diff(X,dims=3)
     @test_throws ArgumentError diff(X,dims=-1)
 end
@@ -2291,6 +2375,8 @@ end
 
     @test accumulate(min, [1, 2, 5, -1, 3, -2]) == [1, 1, 1, -1, -1, -2]
     @test accumulate(max, [1, 2, 5, -1, 3, -2]) == [1, 2, 5, 5, 5, 5]
+    @test Base.accumulate_pairwise(min, [1, 2, 5, -1, 3, -2]) == [1, 1, 1, -1, -1, -2]
+    @test Base.accumulate_pairwise(max, [1, 2, 5, -1, 3, -2]) == [1, 2, 5, 5, 5, 5]
 
     @test accumulate(max, [1 0; 0 1], dims=1) == [1 0; 1 1]
     @test accumulate(max, [1 0; 0 1], dims=2) == [1 1; 0 1]
@@ -2322,6 +2408,9 @@ end
                 @test out ≈ accumulate_arr
             end
         end
+        arr_cop = similar(arr)
+        cumprod!(arr_cop, arr)
+        @test arr_cop ≈ cumprod(arr)
     end
 
     # exotic indexing
@@ -2455,13 +2544,13 @@ end
 
 # Ensure we can hash strange custom structs — and they hash the same in arrays
 struct totally_not_five26034 end
-Base.isequal(::totally_not_five26034, x)=isequal(5,x);
-Base.isequal(x, ::totally_not_five26034)=isequal(5,x);
+Base.isequal(::totally_not_five26034, x::Number)=isequal(5,x);
+Base.isequal(x::Number, ::totally_not_five26034)=isequal(5,x);
 Base.isequal(::totally_not_five26034, ::totally_not_five26034)=true;
 Base.hash(::totally_not_five26034, h::UInt)=hash(5, h);
 import Base.==
-==(::totally_not_five26034, x)= (5==x);
-==(x,::totally_not_five26034)= (5==x);
+==(::totally_not_five26034, x::Number)= (5==x);
+==(x::Number,::totally_not_five26034)= (5==x);
 ==(::totally_not_five26034,::totally_not_five26034)=true;
 @testset "issue #26034" begin
     n5 = totally_not_five26034()
@@ -2519,3 +2608,7 @@ Base.view(::T25958, args...) = args
     @test t[end,end,1]   == @view(t[end,end,1])   == @views t[end,end,1]
     @test t[end,end,end] == @view(t[end,end,end]) == @views t[end,end,end]
 end
+
+# Fix oneunit bug for unitful arrays
+@test oneunit([Second(1) Second(2); Second(3) Second(4)]) == [Second(1) Second(0); Second(0) Second(1)]
+
