@@ -136,6 +136,17 @@ end
     @test !in(100,c)
     @test !in(200,s)
 end
+
+@testset "copy!" begin
+    for S = (Set, BitSet)
+        s = S([1, 2])
+        for a = ([1], UInt[1], [3, 4, 5], UInt[3, 4, 5])
+            @test s === copy!(s, Set(a)) == S(a)
+            @test s === copy!(s, BitSet(a)) == S(a)
+        end
+    end
+end
+
 @testset "sizehint, empty" begin
     s = Set([1])
     @test isequal(sizehint!(s, 10), Set([1]))
@@ -387,6 +398,10 @@ end
     unique!(u)
     @test u == [5,"w","we","r"]
     u = [1,2,5,1,3,2]
+    @test unique!(x -> x ^ 2, [1, -1, 3, -3, 5, -5]) == [1, 3, 5]
+    @test unique!(n -> n % 3, [5, 1, 8, 9, 3, 4, 10, 7, 2, 6]) == [5, 1, 9]
+    @test unique!(iseven, [2, 3, 5, 7, 9]) == [2, 3]
+    @test unique!(x -> x % 2 == 0 ? :even : :odd, [1, 2, 3, 4, 2, 2, 1]) == [1, 2]
 end
 
 @testset "allunique" begin
